/* Crimson Fields - a game of tactical warfare
   Copyright (C) 2000-2007 Jens Granseuer

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

/* SDL_zlib.h */

#ifndef _SDL_ZLIB_H
#define _SDL_ZLIB_H

#include "SDL_rwops.h"

#ifdef __cplusplus
extern "C" {
#endif

SDL_RWops *SDL_RWFromGzip(const char *file, const char *mode);

#ifdef __cplusplus
}
#endif

#endif /* _SDL_ZLIB_H */

