// CoMET - The Crimson Fields Map Editing Tool
// Copyright (C) 2002-2007 Jens Granseuer
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//

///////////////////////////////////////////////////////////////
// msgs.h - message identifiers
///////////////////////////////////////////////////////////////

#ifndef _INCLUDE_ED_MSGS_H
#define _INCLUDE_ED_MSGS_H

#define MSG(m)			m

#define MSG_B_OK		"_OK"
#define MSG_B_CANCEL		"_Cancel"
#define MSG_B_LOAD		"_Load"
#define MSG_B_SAVE		"_Save"
#define MSG_B_SKIP		"_Abort"
#define MSG_LOAD_GAME		"Load game"
#define MSG_SAVE_GAME		"Save game"

#endif   /* _INCLUDE_ED_MSGS_H */

